package org.rakura.rsupdater;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.asm.Opcodes;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.FieldInsnNode;
import org.objectweb.asm.tree.FieldNode;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.InsnNode;
import org.objectweb.asm.tree.MethodNode;
import org.objectweb.asm.tree.VarInsnNode;
import org.rakura.rsupdater.util.TypeQuerier;

public class FieldHolder {

	private static final Map<Character, Integer> TYPE_TO_RETURN = new HashMap<Character, Integer>();

	static {
		TYPE_TO_RETURN.put('L', Opcodes.ARETURN);
		TYPE_TO_RETURN.put('[', Opcodes.ARETURN);
		TYPE_TO_RETURN.put('I', Opcodes.IRETURN);
		TYPE_TO_RETURN.put('Z', Opcodes.IRETURN);
		TYPE_TO_RETURN.put('B', Opcodes.IRETURN);
		TYPE_TO_RETURN.put('C', Opcodes.IRETURN);
		TYPE_TO_RETURN.put('S', Opcodes.IRETURN);
		TYPE_TO_RETURN.put('J', Opcodes.LRETURN);
		TYPE_TO_RETURN.put('F', Opcodes.FRETURN);
		TYPE_TO_RETURN.put('D', Opcodes.DRETURN);
	}

	private final ClassNode fieldOwner;
	private final FieldNode fieldNode;

	@Override
	public boolean equals(Object other) {
		if (other == this)
			return true;
		if (other instanceof FieldHolder) {
			FieldHolder otherFieldHolder = (FieldHolder) other;
			return otherFieldHolder.fieldOwner.name
					.equals(this.fieldOwner.name)
					&& otherFieldHolder.fieldNode.name
							.equals(this.fieldNode.name);
		}
		return false;
	}

	public FieldHolder(ClassNode fieldOwner, FieldNode fieldNode) {
		this.fieldOwner = fieldOwner;
		this.fieldNode = fieldNode;
	}

	@SuppressWarnings("unchecked")
	private static void injectAccessorImpl(ClassNode hostClass,
			ClassNode fieldOwner, FieldNode fieldNode, String realName) {
		String retType = GlobalVariables.INSTANCE
				.getAccessorTypeDesc(fieldNode.desc);
		MethodNode accessor = new MethodNode(Opcodes.ACC_PUBLIC
				| Opcodes.ACC_FINAL, "get"
				+ Character.toUpperCase(realName.charAt(0))
				+ realName.substring(1), "()" + retType, null, null);
		InsnList instructions = new InsnList();
		boolean isInstance = new TypeQuerier(fieldNode.access)
				.isNot(Opcodes.ACC_STATIC);
		if (isInstance)
			instructions.add(new VarInsnNode(Opcodes.ALOAD, 0));
		instructions.add(new FieldInsnNode(isInstance ? Opcodes.GETFIELD
				: Opcodes.GETSTATIC, fieldOwner.name, fieldNode.name,
				fieldNode.desc));
		instructions.add(new InsnNode(TYPE_TO_RETURN.get(fieldNode.desc
				.charAt(0))));
		accessor.instructions = instructions;
		accessor.maxStack = 1;
		accessor.maxLocals = isInstance ? 1 : 0;
		synchronized (hostClass) {
			hostClass.methods.add(accessor);
		}
	}

	public ClassNode getDeclaringClass() {
		return fieldOwner;
	}

	public FieldNode getFieldNode() {
		return fieldNode;
	}

	public void injectAccessorAs(String realName) {
		if (new TypeQuerier(fieldNode.access).is(Opcodes.ACC_STATIC)) {
			injectAccessorImpl(GlobalVariables.INSTANCE.getClientClass(),
					fieldOwner, fieldNode, realName);
		} else {
			injectAccessorImpl(fieldOwner, fieldOwner, fieldNode, realName);
		}
	}

	/**
	 * Can only be used AFTER identifying.
	 */
	public void injectAccessor() {
		injectAccessorAs(GlobalVariables.INSTANCE.getRealName(this));
	}

}