package org.rakura.rsupdater;

import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.CopyOnWriteArrayList;

import org.objectweb.asm.Type;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.FieldNode;
import org.rakura.rsupdater.IdentificationCollisionException.CollisionType;
import org.rakura.rsupdater.identifiers.Identifications;
import org.rakura.rsupdater.identifiers.RSIdentifier;
import org.rakura.rsupdater.transformers.RSTransformer;
import org.rakura.rsupdater.util.Association;

/**
 * Thread safe.
 */
public enum GlobalVariables {

	INSTANCE;

	private volatile boolean canUseClasses = false;

	private volatile ClassNode clientClass;
	private final ConcurrentMap<ClassNode, List<FieldNode>> classToFields = new ConcurrentHashMap<ClassNode, List<FieldNode>>();

	private final ConcurrentMap<ClassNode, Accessor> classToAccessor = new ConcurrentHashMap<ClassNode, Accessor>();
	private final ConcurrentMap<String, Accessor> classNameToAccessor = new ConcurrentHashMap<String, Accessor>();
	private final ConcurrentMap<String, FieldHolder> realNameToField = new ConcurrentHashMap<String, FieldHolder>();

	public static final List<RSIdentifier> IDENTIFIERS = new CopyOnWriteArrayList<RSIdentifier>();
	public static final List<RSTransformer> TRANSFORMERS = new CopyOnWriteArrayList<RSTransformer>();

	static {
		/*
		 * identifiers.add(...); ... transformers.add(...); ...
		 */
	}

	public void init() {
		canUseClasses = true;
	}

	public ClassNode getClientClass() {
		if (!canUseClasses)
			throw new IllegalStateException();
		return clientClass;
	}

	public void setClientClass(ClassNode clientClass) {
		this.clientClass = clientClass;
	}

	public Set<ClassNode> getClasses() {
		if (!canUseClasses)
			throw new IllegalStateException();
		return classToFields.keySet();
	}

	@SuppressWarnings("unchecked")
	public void addClass(ClassNode classNode) {
		classToFields.put(classNode, classNode.fields);
	}

	public Accessor getAccessorTo(ClassNode classNode) {
		return classToAccessor.get(classNode);
	}

	public void mapClassToAccessor(ClassNode classNode, Accessor accessor) {
		Accessor prevValue = classNameToAccessor.put(classNode.name, accessor);
		if (prevValue != null && !(prevValue == accessor))
			throw new IdentificationCollisionException(CollisionType.CLASS,
					classNode.name);
		classToAccessor.put(classNode, accessor);
	}

	public FieldHolder getField(String realName) {
		return realNameToField.get(realName);
	}

	public String getRealName(FieldHolder fieldHolder) {
		Set<String> realNames = realNameToField.keySet();
		for (String realName : realNames)
			if (realNameToField.get(realName).equals(fieldHolder))
				return realName;
		return null;
	}

	public void mapRealNameToField(String realName, FieldHolder field) {
		FieldHolder prevValue = realNameToField.put(realName, field);
		if (prevValue != null && !prevValue.equals(field))
			throw new IdentificationCollisionException(CollisionType.FIELD,
					realName);
	}

	public void submitIdentification(Identifications ident) {
		Map<String, FieldHolder> map = ident.getFieldIdentifications();
		Set<String> realNames = map.keySet();
		for (String realName : realNames)
			if (realNameToField.containsKey(realName))
				throw new IdentificationCollisionException(CollisionType.FIELD,
						realName);
		realNameToField.putAll(map);
		Association<ClassNode, Accessor> classIdentification = ident
				.getClassIdentification();
		mapClassToAccessor(classIdentification.getKey(), classIdentification
				.getValue());
	}

	public String getAccessorTypeDesc(String typeDesc) {
		if (typeDesc.endsWith(";")) {
			Type type = Type.getType(typeDesc);
			int arrayDimensions = 0;
			String internalName = type.getSort() == Type.ARRAY ? Type.getType(
					typeDesc.substring(arrayDimensions = type.getDimensions()))
					.getInternalName() : type.getInternalName();
			String interfaceInternalName = classNameToAccessor
					.get(internalName).getPath();
			if (interfaceInternalName == null)
				return null;
			String interfaceDescriptor = Type.getObjectType(
					interfaceInternalName).getDescriptor();
			if (arrayDimensions == 0)
				return interfaceDescriptor;
			StringBuilder sb = new StringBuilder();
			for (int i = 0; i < arrayDimensions; i++)
				sb.append('[');
			return sb.append(interfaceDescriptor).toString();
		}
		return null;
	}

}