package org.rakura.rsupdater;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.Collections;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarOutputStream;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.IntInsnNode;
import org.objectweb.asm.tree.MethodNode;
import org.rakura.rsupdater.identifiers.Identifications;
import org.rakura.rsupdater.identifiers.RSIdentifier;
import org.rakura.rsupdater.transformers.RSTransformer;

public class Updater implements Runnable {

	private JarFile jar;

	private int rsversion = -1;

	@SuppressWarnings("unchecked")
	public Updater() {
		try {
			byte[] bytes = download(new URL(
					"http://world2.runescape.com/runescape.jar"));
			File runescape = File.createTempFile("runescape", ".jar");
			FileOutputStream out = new FileOutputStream(runescape);
			out.write(bytes);
			out.close();
			bytes = null;

			System.gc();

			jar = new JarFile(runescape);

			Enumeration<JarEntry> en = jar.entries();
			while (en.hasMoreElements()) {
				JarEntry entry = en.nextElement();
				if (entry.getName().endsWith(".class")) {
					ClassNode cn = new ClassNode();
					ClassReader cr = new ClassReader(jar.getInputStream(entry));
					cr.accept(cn, 0);
					GlobalVariables.INSTANCE.addClass(cn);
					if (cn.name.equals("client")) {
						GlobalVariables.INSTANCE.setClientClass(cn);
						List<MethodNode> methodNodes = cn.methods;
						MethodNode mainMethodNode = null;
						for (MethodNode methodNode : methodNodes) {
							if (methodNode.name.equals("main")) {
								mainMethodNode = methodNode;
								break;
							}
						}
						if (mainMethodNode != null) {
							InsnList instructions = mainMethodNode.instructions;
							for (Iterator<AbstractInsnNode> itr = instructions
									.iterator(); itr.hasNext();) {
								AbstractInsnNode nextNode = itr.next();
								if (nextNode.getOpcode() == Opcodes.SIPUSH) {
									int operand = ((IntInsnNode) nextNode).operand;
									if (operand < 768 && operand > 0) {
										rsversion = operand;
										break;
									}
								}
							}
						} else
							throw new AssertionError(
									"No main method in client.class");
					}
				}
			}

			jar.close();

			GlobalVariables.INSTANCE.init();

		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	/**
	 * @author Kosaki
	 * 
	 */
	private static byte[] download(URL url) throws IOException {
		URLConnection uc = url.openConnection();
		int len = uc.getContentLength();
		InputStream is = new BufferedInputStream(uc.getInputStream());
		try {
			byte[] data = new byte[len];
			int offset = 0;
			while (offset < len) {
				int read = is.read(data, offset, data.length - offset);
				if (read < 0) {
					break;
				}
				offset += read;
			}
			if (offset != len) {
				throw new IOException(String.format(
						"Read %d bytes; Expected %d bytes", offset, len));
			}
			return data;
		} finally {
			is.close();
		}
	}

	@Override
	public void run() {
		try {
			long startMillis = System.currentTimeMillis();
			System.out.printf(
					"Updater log for RS-%d: %2$tB %2$te, %2$tY %2$tr%n",
					rsversion, new Date());
			Map<String, byte[]> name2Bytes = new HashMap<String, byte[]>();
			List<RSIdentifier> identifiers = GlobalVariables.IDENTIFIERS;
			int numberOfIdentifications = identifiers.size();
			ExecutorService executor = Executors
					.newFixedThreadPool(numberOfIdentifications);
			ExecutorCompletionService<Identifications> completionService = new ExecutorCompletionService<Identifications>(
					executor);
			Collections.sort(identifiers);
			for (RSIdentifier identifier : identifiers)
				completionService.submit(identifier);
			for (int i = 0; i < numberOfIdentifications; ++i)
				GlobalVariables.INSTANCE.submitIdentification(completionService
						.take().get());
			List<RSTransformer> transformers = GlobalVariables.TRANSFORMERS;
			int numberOfTransformations = transformers.size();
			for (RSTransformer transformer : transformers)
				completionService.submit(transformer, null);
			for (int i = 0; i < numberOfTransformations; ++i)
				completionService.take();
			Set<ClassNode> classes = GlobalVariables.INSTANCE.getClasses();
			for (ClassNode classNode : classes) {
				ClassWriter classWriter = new ClassWriter(0);
				classNode.accept(classWriter);
				name2Bytes.put(classNode.name, classWriter.toByteArray());
			}
			Set<Entry<String, byte[]>> entries = name2Bytes.entrySet();
			String fileName = System.getProperty("user.home") + File.separator
					+ "runescape_" + rsversion + "_injected.jar";
			File file = new File(fileName);
			FileOutputStream fos = new FileOutputStream(file);
			JarOutputStream stream = new JarOutputStream(fos);
			for (Entry<String, byte[]> entry : entries) {
				JarEntry je = new JarEntry(entry.getKey() + ".class");
				stream.putNextEntry(je);
				stream.write(entry.getValue());
			}
			stream.close();
			System.out.printf(
					"Successfully injected to: %s in %d milliseconds",
					fileName, System.currentTimeMillis() - startMillis);
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

}