package org.rakura.rsupdater.identifiers;

import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.CopyOnWriteArraySet;

import org.objectweb.asm.tree.ClassNode;
import org.rakura.rsupdater.GlobalVariables;

public abstract class RSIdentifier implements Callable<Identifications>,
		Comparable<RSIdentifier> {

	protected static final Set<ClassNode> CLASSES = new CopyOnWriteArraySet<ClassNode>(
			GlobalVariables.INSTANCE.getClasses());

	public final RSIdentifierManifest ANNOTATION = getClass().getAnnotation(
			RSIdentifierManifest.class);

	public boolean isDependentUpon(RSIdentifier other) {
		if (ANNOTATION == null)
			return false;
		Class<? extends RSIdentifier>[] dependencies = ANNOTATION
				.dependencies();
		for (Class<? extends RSIdentifier> dependency : dependencies)
			if (dependency.getName().equals(other.getClass().getName()))
				return true;
		return false;
	}

	@Override
	public int compareTo(RSIdentifier other) {
		if (other.isDependentUpon(this))
			return -1;
		else if (this.isDependentUpon(other))
			return 1;
		return 0;
	}

}